"""
PDFMiner.six compatibility, which now just pulls stuff from playa.miner.

This will definitely go away in PAVÉS 1.0.
"""

from playa.miner import (
    ColorSpace,
    GraphicState,
    KWD,
    LIT,
    NameTree,
    NumberTree,
    PDFDocEncoding,
    PDFDocument,
    PDFPage,
    PDFException,
    PDFObjRef,
    PDFObject,
    PSException,
    PSLiteral,
    decode_text,
    pdftypes,
    resolve1,
    resolve_all,
    PathSegment,
    PDFTypeError,
    PDFValueError,
    Plane,
    LAParams,
    LTItem,
    LTText,
    LTComponent,
    LTCurve,
    LTLine,
    LTRect,
    LTImage,
    LTAnno,
    LTChar,
    LTContainer,
    LTExpandableContainer,
    LTTextContainer,
    TextLineElement,
    LTTextLine,
    LTTextLineHorizontal,
    LTTextLineVertical,
    LTTextBox,
    LTTextBoxHorizontal,
    LTTextBoxVertical,
    TextGroupElement,
    LTTextGroup,
    LTTextGroupLRTB,
    LTTextGroupTBRL,
    LTLayoutContainer,
    LTFigure,
    LTPage,
    extract_page,
    extract,
)


__all__ = [
    "ColorSpace",
    "GraphicState",
    "KWD",
    "LIT",
    "NameTree",
    "NumberTree",
    "PDFDocEncoding",
    "PDFDocument",
    "PDFPage",
    "PDFException",
    "PDFObjRef",
    "PDFObject",
    "PSException",
    "PSLiteral",
    "decode_text",
    "extract",
    "extract_page",
    "pdftypes",
    "resolve1",
    "resolve_all",
    "PathSegment",
    "PDFTypeError",
    "PDFValueError",
    "Plane",
    "LAParams",
    "LTItem",
    "LTText",
    "LTComponent",
    "LTCurve",
    "LTLine",
    "LTRect",
    "LTImage",
    "LTAnno",
    "LTChar",
    "LTContainer",
    "LTExpandableContainer",
    "LTTextContainer",
    "TextLineElement",
    "LTTextLine",
    "LTTextLineHorizontal",
    "LTTextLineVertical",
    "LTTextBox",
    "LTTextBoxHorizontal",
    "LTTextBoxVertical",
    "TextGroupElement",
    "LTTextGroup",
    "LTTextGroupLRTB",
    "LTTextGroupTBRL",
    "LTLayoutContainer",
    "LTFigure",
    "LTPage",
]
